# ==============================================================
#  RobotVision@ImageCLEF'12 Point cloud library images generator
#
#  2012-03-28
#
#
#  Author: Jesus Martinez
#                 Jesus.Martinez@uclm.es      http://neithan.weebly.com/
#                 
# ==============================================================   

# --------------------------------------------------------------
# Imports
# --------------------------------------------------------------

import sys
import os
import re
from pylab import *
from PIL import Image


minDistance = -10
scaleFactor = 0.0021

# --------------------------------------------------------------
# Depth2Cloud
# --------------------------------------------------------------

# Generates an image compatible with the point cloud library (.pcd) from the depth images generated with the kinect device
#
# For more information realted to the .pcd formate, please visit http://pointclouds.org/documentation/tutorials/pcd_file_format.php
#
# Parameters:
#
# - input_depth_image - Path to the input .jpg file
# - output_cloud_image - Path to the output .pcd file
#
#

# --------------------------------------------------------------
# Main
# --------------------------------------------------------------


def main():
  cmd = CmdLine()
  cmd.setAppAuthor("Jesus Martinez-Gomez")
  cmd.setAppName("depth2cloud.py")
  cmd.setAppDescr("RobotVision@ImageCLEF'12 Point Cloud Library Images Generator Script")
  # Arguments
  cmd.addStrArg("input_depth_image", "Path to the input file. This file should be a .jpg depth image from the RobotVision'12 datasets")
  cmd.addStrArg("output_cloud_image", "Path to the output file. This file should be a .pcd file that will be created or overwritten")
  
  # Header
  cmd.printHeader()
  # Parse command line arguments and display help
  if not cmd.parse():
    cmd.printHelp()
    sys.exit(1)
  # Print selected args
  cmd.printValues()
    
  
  imgfile = cmd.arg_input_depth_image
  cloudfile = cmd.arg_output_cloud_image
  
  print 'Reading colour image...\n'
    
  try:    
    img = Image.open(imgfile)
    w, h = img.size
    print "Input image information: "+str(w)+" in width and "+str(h)+" in heigh\n"
    
  except Exception, e:
    print "ERROR: " + str(e)    
    print "ERROR: Unable to open the input file'"+cmd.arg_input_depth_image+"'!\n"
    sys.exit(1)  
  
  print 'Generating cloud file...\n'

  points = 0

  # Compute the number of points that will be generated
  
  for i in range(img.size[0]):
      for j in range(img.size[1]):
        
          r,g,b = img.getpixel((i,j))
          
          condr = 0
          condg = 0
          condb = 0
          
          if(r==255 and (g==0 or b==0)):
            condr = 1
          if(g==255 and (r==0 or b==0)):
            condg = 1
          if(b==255 and (r==0 or g==0)):
            condb = 1
            
          if (condr==1 or condg==1 or condb==1):

            r = float(r)
            g = float(g)
            b = float(b)
            newPoint = 0

            if r==255 and g!=255 and b==0 and newPoint==0:
              # R=255 - G is used for distance - B=0
              newPoint = 1
              points=points+1

            elif r==255 and g==255 and b==0 and newPoint==0:
              # mix between distance 1 and 2
              newPoint = 1
              points=points+1             

            elif r!=255 and g==255 and b==0 and newPoint==0:
              # R is used for distance G=255 B=0
              newPoint = 1
              points=points+1

            elif r==0 and g==255 and b!=255 and newPoint==0:
              # B is used for distance G=255 R=0
              newPoint = 1
              points=points+1

            elif r==0 and g==255 and b==255 and newPoint==0:
              # mix between distance 3 and 4
              newPoint = 1
              points=points+1             
              
            elif r==0 and g!=255 and b==255 and newPoint==0:
              # G is used for distance B=255 R=0
              newPoint = 1
              points=points+1
            else:
              print r,g,b

  try:    
    pcd = open(cloudfile, 'wt')

    # Default values for the .pcd header
    
    pcd.write('VERSION .7\n')
    pcd.write('FIELDS x y z rgb\n')
    pcd.write('SIZE 4 4 4 4\n')
    pcd.write('TYPE F F F F\n')
    pcd.write('COUNT 1 1 1 1\n')
    pcd.write('WIDTH %d\n' % points)
    pcd.write('HEIGHT 1\n')
    pcd.write('VIEWPOINT 0 0 0 1 0 0 0\n')
    pcd.write('POINTS %d\n' %points)
    pcd.write('DATA ascii\n')
    
  except Exception, e:
    print "ERROR: " + str(e)    
    print "ERROR: Unable to open the output file'"+cmd.arg_output_cloud_image+"'!\n"
    sys.exit(1)
  
  initDist = 0.4
  widthDist1 = 0.8
  widthDist2 = 0.8
  widthDist3 = 0.8
  widthDist4 = 0.8

  points2 = 0

  weightMinVar = 0

  # Convert from pixel ref (i, j, z) to 3D space (x,y,z)
  for i in range(img.size[0]):
      for j in range(img.size[1]):
          
          z=0
          r,g,b = img.getpixel((i,j))
          
          condr = 0
          condg = 0
          condb = 0
          
          if(r==255 and (g==0 or b==0)):
            condr = 1
          if(g==255 and (r==0 or b==0)):
            condg = 1
          if(b==255 and (r==0 or g==0)):
            condb = 1
            
          if (condr==1 or condg==1 or condb==1):
            i = float(i)
            j = float(j)


            x = float(i/img.size[0]) - 0.5
            y = float(j/img.size[1]) - 0.5
            x = x*(img.size[0]/img.size[1])

            r = float(r)
            g = float(g)
            b = float(b)

            rTemp=0
            gTemp=0
            bTemp=0

            newPoint = 0

            if r==255 and g!=255 and b==0 and newPoint==0 :
              
              # R=255 - G is used for distance - B=0
              gTemp = max(weightMinVar,g) # 100 - 255
              gTemp = gTemp - weightMinVar    # 0 - 155

              weight = float(gTemp/(255-weightMinVar))
              
              z = initDist + weight*widthDist1
              pcd.write("%f  %f  %f 0.2108e+06\n" % (x, y, z))
              points2=points2+1
              newPoint = 1

            if r==255 and g==255 and b==0 and newPoint==0 :
              # mix between distance 1 and 2              
              gTemp = 255 - weightMinVar    # 0 - 155

              weight = float(gTemp/(255-weightMinVar))
              
              z = initDist + weight*widthDist1
              pcd.write("%f  %f  %f 0.2108e+06\n" % (x, y, z))
              points2=points2+1
              newPoint = 1

            if g==255 and b==0 and r!=255 and newPoint==0:
              # R is used for distance G=255 B=0
              rTemp = max(weightMinVar,r) # 100 - 255
              rTemp = rTemp - weightMinVar    # 0 - 155
              
              weight = float(1- float(rTemp/(255-weightMinVar)))
              
              z = initDist + widthDist1 + float(weight*widthDist2)
              pcd.write("%f  %f  %f 2.2108e+06\n" % (x, y, z))
              points2=points2+1
              newPoint = 1

            if g==255 and r==0 and b!=255 and newPoint==0:
              # B is used for distance G=255 R=0
              bTemp = max(weightMinVar,b) # 100 - 255
              bTemp = bTemp - weightMinVar    # 0 - 155
              
              weight = float(bTemp/(255-weightMinVar))
              
              z = initDist + widthDist1 + widthDist2 + float(weight*widthDist3)
              pcd.write("%f  %f  %f 4.2108e+06\n" % (x, y, z))
              points2=points2+1
              newPoint = 1

            if r==0 and g==255 and b==255 and newPoint==0 :
              # mix between distance 3 and 5                            
              bTemp = 255 - weightMinVar    # 0 - 155              
              weight = float(bTemp/(255-weightMinVar))
              
              z = initDist + widthDist1 + widthDist2 + float(weight*widthDist3)
              pcd.write("%f  %f  %f 4.2108e+06\n" % (x, y, z))
              points2=points2+1
              newPoint = 1

            if b==255 and r==0 and g!=255 and newPoint==0:
              # G is used for distance B=255 R=0
              gTemp = max(weightMinVar,g) # 100 - 255
              gTemp = gTemp - weightMinVar    # 0 - 155
              
              weight = float(1-float(gTemp/(255-weightMinVar)))
              
              z = initDist + widthDist1 + widthDist2 + widthDist3 + float(weight*widthDist4)
              pcd.write("%f  %f  %f 4.2108e+06\n" % (x, y, z))
              points2=points2+1
              newPoint = 1
              
  pcd.close
  print 'points 1 %d' % points
  print 'points 2 %d' % points2
  print 'Done!'

# --------------------------------------------------------------
# Internal
# --------------------------------------------------------------
def getLines(string):
        return string.split('\n')

def printNoNewLine(str):
        sys.stdout.write(str)
        sys.stdout.flush()

def strPadRight(string, num, char=" "):
        for i in range(num):
                string+=char
        return string


def strCenter(string, width, char=" "):
        left=True
        while(len(string)<width):
                if left:
                        string=char+string;
                        left=False
                else:
                        string+=char;
                        left=True
        return string

class CmdLine:
        def __init__(self):
                self.argv = sys.argv
                self.arguments = list()
                self.appName=os.path.basename(sys.argv[0])
                self.appDescr=""
                self.appAuthor=""
                self.options = list()

        def addOpt(self, shrt, lng="", description="", typ=bool(), default=False, argName=""):
                for o in self.options:
                        if len(shrt)>0:
                                assert (o[0]!=shrt)
                        if len(lng)>0:
                                assert (o[1]!=lng)
                self.options.append((shrt, lng, description, typ, argName))
                exec("self.opt_"+lng+"=default")

        def addBoolOpt(self, shrt, lng="", description="", default=False):
                self.addOpt(shrt, lng, description, bool(), default)

        def addIntOpt(self, shrt, lng="", description="", default=0, argName="int"):
                self.addOpt(shrt, lng, description, int(), default, argName)

        def addStrOpt(self, shrt, lng="", description="", default="", argName="str"):
                self.addOpt(shrt, lng, description, str(), default, argName)

        def addArg(self, name, description="", typ=str(), values=[]):
                for a in self.arguments:
                        assert (a[0]!=name)
                self.arguments.append((name, description, typ, values))

        def addStrArg(self, name, description=""):
                self.addArg(name, description, str())

        def addBoolArg(self, name, description=""):
                self.addArg(name, description, bool())

        def addIntArg(self, name, description=""):
                self.addArg(name, description, int())

        def addFloatArg(self, name, description=""):
                self.addArg(name, description, float())

        def addEnumArg(self, name, description, values):
                for i in range(len(values)):
                        values[i]=str(values[i])
                self.addArg(name, description, str(), values)

        def setAppName(self, name):
                self.appName=name

        def setAppDescr(self, descr):
                self.appDescr=descr

        def setAppAuthor(self, author):
                self.appAuthor=author

        def printError(self, e):
                print "Error: "+e+"\n"

        def printHeader(self):
                # Create strings
                authorStr = "Author: "+self.appAuthor
                nameStr = self.appName
                descrStr = getLines(self.appDescr)
                # Find max length
                maxLen=0;
                if len(authorStr)>maxLen:
                        maxLen=len(authorStr)
                if len(nameStr)>maxLen:
                        maxLen=len(nameStr)
                for i in descrStr:
                        if len(i)>maxLen:
                                maxLen=len(i)
                # Display
                line1=strPadRight(str(),maxLen+2, "=")
                line2=strPadRight(str(),maxLen+2, "-")
                print "/"+line1+"\\"
                print "|"+strCenter(nameStr, maxLen+2)+"|"
                print "|"+line2+"|"
                for d in descrStr:
                        print "| "+strPadRight(d, maxLen-len(d)+1)+"|"
                print "| "+strPadRight(authorStr, maxLen-len(authorStr)+1)+"|"
                print "\\"+line1+"/"
                print

        def getOptStr(self, o):
                optStr=""
                if len(o[0])>0:
                        optStr+="-"+o[0]
                        if len(o[1])>0:
                                optStr+=", "
                if len(o[1])>0:
                        optStr+="--"+o[1]
                return optStr   

        def printHelp(self):
                # Usage
                usageStr="Usage: "+self.argv[0]
                printNoNewLine(usageStr)
                # Argument list
                maxLen=0
                usageLen=0
                if len(self.options)>0:
                        opStr=" [Options]"
                        printNoNewLine(opStr)
                        usageLen+=len(opStr)
                for i in self.arguments:
                        argStr=" <"+i[0]+">"
                        if usageLen+len(argStr)>80:
                                print ""
                                printNoNewLine(strPadRight(str(), len(usageStr)))
                                usageLen=0
                        printNoNewLine(argStr)
                        usageLen+=len(argStr)
                        if len(i[0])>maxLen:
                                maxLen=len(i[0])
                print
                print
                # Argument description
                print "Arguments:"
                emptyStr=strPadRight(str(), maxLen+4)
                for i in self.arguments:
                        n="  <"+i[0]+">"
                        paddedN = strPadRight(n, maxLen+4-len(n))
                        descr = getLines(i[1])
                        print paddedN+" - "+descr[0]
                        for j in descr[1:]:
                                print emptyStr+"   "+j
                print 
                # Option description
                if len(self.options)>0:
                        print "Options:"
                        maxLen=0
                        for o in self.options:
                                if type(o[3])==type(bool()):
                                        optStr="  "+self.getOptStr(o)
                                else:
                                        optStr="  "+self.getOptStr(o)+" <"+o[4]+">"
                                if len(optStr)>maxLen:
                                        maxLen=len(optStr)
                        for o in self.options:
                                if type(o[3])==type(bool()):
                                        optStr="  "+self.getOptStr(o)
                                else:
                                        optStr="  "+self.getOptStr(o)+" <"+o[4]+">"
                                descr = getLines(o[2])
                                print strPadRight(optStr, maxLen-len(optStr))+" - "+descr[0]
                                for d in descr[1:]:
                                        print strPadRight(str(), maxLen)+"   "+d                    

        def parse(self):
                # Check number of arguments
                if len(self.argv)-1 < len(self.arguments):
                        self.printError("Incorrect command line arguments.")
                        return False
                # Decode options
                optCount=0
                endOpt=False
                while (not endOpt) and ((len(self.argv)-1)>optCount):
                        endOpt=True
                        val=self.argv[1+optCount]
                        for o in self.options:
                                shrt=o[0]
                                lng=o[1]
                                typ=o[3]
                                if (val==("-"+shrt)) or (val==("--"+lng)):
                                        endOpt=False
                                        optCount+=1
                                        if type(typ)==type(bool()):
                                                if len(shrt)>0:
                                                        exec("self.opt_"+shrt+"=True")
                                                if len(lng)>0:
                                                        exec("self.opt_"+lng+"=True")
                                        else:
                                                if len(self.argv)-1<optCount+1:
                                                        self.printError("Missing argument of the option "+val+".")
                                                        return False
                                                arg=self.argv[1+optCount]
                                                optCount+=1
                                                try:
                                                        # Convert
                                                        if type(typ)==type(int()):
                                                                arg=int(arg)
                                                        if type(typ)==type(float()):
                                                                arg=float(arg)
                                                except:
                                                        self.printError("Incorrect value of argument of the option "+val+".")
                                                        return False
                                                if len(shrt)>0:
                                                        exec("self.opt_"+shrt+"=arg")
                                                if len(lng)>0:
                                                        exec("self.opt_"+lng+"=arg")
                                        break

                # Decode arguments
                for i in range(len(self.arguments)):
                        name=self.arguments[i][0]
                        typ=self.arguments[i][2]
                        val=self.argv[i+1+optCount]
                        values=self.arguments[i][3]
                        try:
                                # Convert
                                if type(typ)==type(int()):
                                        val=int(val)
                                if type(typ)==type(float()):
                                        val=float(val)
                                if type(typ)==type(bool()):
                                        val=bool(val)
                                # Check values
                                if len(values)>0:
                                        assert val in values
                        except:
                                self.printError("Incorrect value of argument <"+name+">.")
                                return False
        
                        # Create variable
                        exec("self.arg_"+name+"=val")
                # All ok
                return True

        def printValues(self):
                if len(self.arguments)>0:
                        print "Selected Arguments:"
                        maxLen=0
                        for i in self.arguments:
                                if len(i[0])>maxLen:
                                        maxLen=len(i[0])
                        for i in range(len(self.arguments)):
                                name=self.arguments[i][0]
                                val=eval("self.arg_"+name)
                                name="<"+name+">"
                                print "  "+strPadRight(name, maxLen-len(name)+2)+" = "+str(val)
                        print
                if len(self.options)>0:
                        print "Selected Options:"
                        maxLen=0
                        for o in self.options:
                                optStr="  "+self.getOptStr(o)
                                if len(optStr)>maxLen:
                                        maxLen=len(optStr)
                        for o in self.options:
                                optStr="  "+self.getOptStr(o)
                                val=eval("self.opt_"+o[1])
                                print strPadRight(optStr, maxLen-len(optStr))+" = "+str(val)
                        print

# --------------------------------------------------------------
# Execute the main function when run as a script
# --------------------------------------------------------------
if __name__ == "__main__":
        main()
